"""Constants for LetMeKnow."""

from __future__ import annotations

from importlib import metadata
import logging
from typing import Final

LOGGER = logging.getLogger(__package__)
VERSION: Final[str] = metadata.version(__package__)

TARGETS_ALL: Final[list[str]] = []
TARGETS_ALL_CLIENTS: Final[list[str]] = ["client-*"]
TARGETS_ALL_HEADLESS: Final[list[str]] = ["headless-*"]



"""Python client for LetMeKnow."""

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any, Self


class LMKClientType(StrEnum):
    """Enum of client type."""

    CLIENT = "client"
    HEADLESS = "headless"


@dataclass(slots=True)
class LMKNotification:
    """Notification."""

    type: str | None = None
    title: str | None = None
    subtitle: str | None = None
    content: str | None = None
    image: LMKNotificationImage | None = None
    timeout: int | None = None

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            type=result.get("type"),
            title=result.get("title"),
            subtitle=result.get("subtitle"),
            content=result.get("content"),
            image=LMKNotificationImage.from_dict(result["image"])
            if result.get("image")
            else None,
            timeout=result.get("timeout"),
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "type": str(self.type),
            "title": self.title,
            "subtitle": self.subtitle,
            "content": self.content,
            "image": self.image.to_dict() if self.image else None,
            "timeout": self.timeout,
        }


@dataclass(slots=True)
class LMKNotificationImage:
    """Notification image."""

    url: str

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            url=result["url"],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "url": self.url,
        }


class LMKWSRequestType(StrEnum):
    """Enum of websocket request type."""

    REGISTER = "register"
    NOTIFICATION = "notification"


class LMKWSResponseType(StrEnum):
    """Enum of websocket response type."""

    ERROR = "error"
    NOTIFICATION_SENT = "notificationSent"
    REGISTER = "register"
    SUCCESS = "success"  # Generic success response, not sent by the server


@dataclass(slots=True)
class LMKWSRegister:
    """Websocket register client."""

    type: LMKWSRequestType
    user_id: str

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            type=LMKWSRequestType(result["type"]),
            user_id=result["userID"],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "type": str(self.type),
            "userID": self.user_id,
        }


@dataclass(slots=True)
class LMKWSNotification:
    """Websocket notification."""

    type: LMKWSRequestType
    data: LMKNotification
    targets: list[str]

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            type=LMKWSRequestType(result["type"]),
            data=LMKNotification.from_dict(result["data"]),
            targets=result["targets"],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "type": str(self.type),
            "data": self.data.to_dict(),
            "targets": self.targets,
        }


@dataclass(slots=True)
class LMKWSResponseError:
    """Websocket response error."""

    type: LMKWSResponseType
    message: str
    error: str

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            type=LMKWSResponseType(result["type"]),
            message=result["message"],
            error=result["error"],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "type": str(self.type),
            "message": self.message,
            "error": self.error,
        }


@dataclass(slots=True)
class LMKWSResponseSuccess:
    """Websocket response success."""

    type: LMKWSResponseType
    succeeded: bool
    message: str

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            type=LMKWSResponseType(result["type"]),
            succeeded=result["succeeded"],
            message=result["message"],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert class to a dict."""
        return {
            "type": str(self.type),
            "succeeded": self.succeeded,
            "message": self.message,
        }



"""Exceptions for LetMeKnow."""

from __future__ import annotations


class LMKError(Exception):
    """LetMeKnow Generic exception."""

    def __init__(self, message: str | None = None) -> None:
        """Initialize the exception."""
        super().__init__(message or "An unknown error occurred")


class LMKConnectionError(LMKError):
    """LetMeKnow connection exception."""

    def __init__(
        self,
        connection_type: str = "WebSocket",
    ) -> None:
        """Initialize the exception."""
        super().__init__(f"Error connecting to {connection_type}")


class LMKNotConnectedError(LMKError):
    """LetMeKnow not connected exception."""

    def __init__(
        self,
        connection_type: str = "WebSocket",
    ) -> None:
        """Initialize the exception."""
        super().__init__(f"{connection_type} connection not established")



"""Client for communication with LetMeKnow."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from socket import gaierror
from typing import TYPE_CHECKING, Any

from aiohttp import (
    ClientConnectionError,
    ClientSession,
    ClientWebSocketResponse,
    WSMsgType,
    WSServerHandshakeError,
)
from yarl import URL

from .const import LOGGER, TARGETS_ALL_CLIENTS, VERSION
from .exceptions import LMKConnectionError, LMKNotConnectedError
from .models import (
    LMKClientType,
    LMKNotification,
    LMKWSNotification,
    LMKWSRegister,
    LMKWSRequestType,
    LMKWSResponseError,
    LMKWSResponseSuccess,
    LMKWSResponseType,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import Self


@dataclass
class LMKClient:
    """Client to communicate with LetMeKnow."""

    lmk_host: str
    lmk_port: int
    lmk_client_type: LMKClientType
    lmk_user_id: str

    session: ClientSession | None = None
    request_timeout: int = 10
    _close_session: bool = False
    _ws: ClientWebSocketResponse | None = None

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The LMKClient object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self._close()

    async def _close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def _ws_send(
        self,
        data: dict[str, Any],
        wait_for_response: bool | None = None,
    ) -> LMKWSResponseSuccess | LMKWSResponseError:
        """Send data to the websocket server.

        Args:
        ----
            data: Data to send.
            wait_for_response: Wait for a response. Defaults to True.

        """
        if self._ws is None or self._ws.closed:
            raise LMKNotConnectedError

        # Default to waiting for a response
        if wait_for_response is None:
            wait_for_response = True

        LOGGER.debug("Sending data to websocket server: %s", data)

        try:
            await self._ws.send_json(data)
        except (
            ClientConnectionError,
            ConnectionResetError,
            gaierror,
        ) as error:
            raise LMKConnectionError from error

        if not wait_for_response:
            return LMKWSResponseSuccess(
                type=LMKWSResponseType.SUCCESS,
                succeeded=True,
                message="No response expected",
            )

        async with asyncio.timeout(self.request_timeout):
            response = await self._ws.receive()

        if response.type == WSMsgType.TEXT:
            # Get the response data
            response_data: dict[str, Any] = response.json()

            if "error" in response_data:
                return LMKWSResponseError.from_dict(response_data)

            return LMKWSResponseSuccess.from_dict(response_data)

        if response.type in (
            WSMsgType.CLOSE,
            WSMsgType.CLOSED,
            WSMsgType.CLOSING,
        ):
            self._ws = None
            return LMKWSResponseError(
                type=LMKWSResponseType.ERROR,
                message="Connection closed",
                error="Connection closed",
            )

        return LMKWSResponseError(
            type=LMKWSResponseType.ERROR,
            message="Unused response type",
            error=response.type.name,
        )

    @property
    def ws_connected(self) -> bool:
        """Check if the websocket is connected.

        Returns
        -------
            True if the websocket is connected.

        """
        return self._ws is not None and not self._ws.closed

    async def ws_connect(self) -> bool:
        """Connect to LetMeKnow websocket server.

        Returns
        -------
            True if the connection was successful.

        """
        url = URL.build(
            scheme="ws",
            host=self.lmk_host,
            port=self.lmk_port,
        ).joinpath("websocket")

        LOGGER.debug("Connecting to websocket server: %s", url)

        headers = {
            "User-Agent": f"LMKClientPy/{VERSION}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                self._ws = await self.session.ws_connect(
                    url=url,
                    headers=headers,
                    heartbeat=30,
                )
        except (
            asyncio.TimeoutError,
            WSServerHandshakeError,
            ClientConnectionError,
            gaierror,
        ) as error:
            raise LMKConnectionError from error

        return True

    async def ws_keep_alive(self) -> None:
        """Keep the websocket connection alive."""
        while self.ws_connected:
            await asyncio.sleep(10)

        LOGGER.info("Websocket connection closed, reconnecting")

        if not await self.ws_connect():
            LOGGER.error("Failed to reconnect to websocket server")
            return

        if (
            register_response := await self.ws_register()
        ).type != LMKWSResponseType.REGISTER:
            LOGGER.error(
                "Failed to register with websocket server: %s", register_response
            )
            return

        LOGGER.info("Reconnected to websocket server")

        # Continue to keep the connection alive
        await self.ws_keep_alive()

    async def ws_register(self) -> LMKWSResponseSuccess | LMKWSResponseError:
        """Register with the websocket server.

        Returns
        -------
            The response from the websocket server.

        """
        LOGGER.debug("Registering with the websocket server as: %s", self.lmk_user_id)

        return await self._ws_send(
            LMKWSRegister(
                type=LMKWSRequestType.REGISTER,
                user_id=self.lmk_user_id,
            ).to_dict()
        )

    async def ws_send_notification(
        self,
        notification: LMKNotification,
        targets: list[str] | None = None,
    ) -> LMKWSResponseSuccess | LMKWSResponseError:
        """Send a notification to the websocket server.

        Args:
        ----
            notification: Notification to send.
            targets: List of targets to send to. Defaults to all clients.

        Returns:
        -------
            The response from the websocket server.

        """
        if targets is None:
            targets = TARGETS_ALL_CLIENTS

        LOGGER.debug("Sending notification to the websocket server: %s", notification)
        LOGGER.debug("Notification targets: %s", targets)

        return await self._ws_send(
            LMKWSNotification(
                type=LMKWSRequestType.NOTIFICATION,
                data=notification,
                targets=targets,
            ).to_dict()
        )

    async def ws_listen_for_notifications(
        self,
        cb: Callable[[LMKNotification], None],
    ) -> None:
        """Listen for notifications.

        Args:
        ----
            cb: Callback to call when a notification is received.

        """
        if self._ws is None or self._ws.closed:
            raise LMKNotConnectedError

        async for message in self._ws:
            if message.type == 1:
                response_data: dict[str, Any] = message.json()
                notification = LMKWSNotification.from_dict(response_data)
                cb(notification.data)
            else:
                break



"""Client for LetMeKnow."""

from .const import TARGETS_ALL, TARGETS_ALL_CLIENTS, TARGETS_ALL_HEADLESS, VERSION
from .exceptions import LMKConnectionError, LMKError, LMKNotConnectedError
from .lmk import LMKClient
from .models import (
    LMKClientType,
    LMKNotification,
    LMKNotificationImage,
    LMKWSNotification,
    LMKWSRegister,
    LMKWSRequestType,
    LMKWSResponseError,
    LMKWSResponseSuccess,
    LMKWSResponseType,
)
from .utils import generate_user_id

__all__ = [
    "TARGETS_ALL",
    "TARGETS_ALL_CLIENTS",
    "TARGETS_ALL_HEADLESS",
    "VERSION",
    "LMKClient",
    "LMKConnectionError",
    "LMKError",
    "LMKNotConnectedError",
    "LMKClientType",
    "LMKNotification",
    "LMKNotificationImage",
    "LMKWSNotification",
    "LMKWSRegister",
    "LMKWSRequestType",
    "LMKWSResponseError",
    "LMKWSResponseSuccess",
    "LMKWSResponseType",
    "generate_user_id",
]



"""Utility functions for the LetMeKnow."""

from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import uuid4

if TYPE_CHECKING:
    from .models import LMKClientType


def generate_user_id(client_type: LMKClientType) -> str:
    """Generate a user ID."""
    return f"{client_type}-{uuid4()!s}"
